/*
ȉ̃TCgŔzzĂDCT-iDCT̃\[XςĎgpĂ܂B

Ooura's Mathematical Software Packages
http://www.kurims.kyoto-u.ac.jp/~ooura/index-j.html

ύX_Ƃ܂Ă

E16*16DCT-iDCT֐
E񎟌zꎟzɕύX
E_gpĂ̂𐮐݂̂

łB
̕ύX_ŁAŌ̂̂łˁcc̃\[XɂȂΕ邱ƐȂ̂łA蔲A̈@ŕύXĂ܂BƍHvAƎłȂ̂łBł̗͂URTCϊłxȂ邾낤ƎvAn_shift 1 ɂČׂĂ݂ĂA܂A_}[ϊǂȂ̂łˁBȂ̂ŁÂ܂܂Ŏ܂B
10rbgΉSIMDgpDCT-iDCT֐͂ōΗǂ̂łcc܂͒w̐w畜K܂ƁB
*/

/*
Short Discrete Cosine Transform
    data length :8x8, 16x16
    method      :row-column, radix 4 FFT
functions
    ddct8x8s  : 8x8 DCT
    ddct16x16s: 16x16 DCT
function prototypes
    void ddct8x8s(int isgn, double **a);
    void ddct16x16s(int isgn, double **a);
*/

/*
-------- 8x8 DCT (Discrete Cosine Transform) / Inverse of DCT --------
    [definition]
        <case1> Normalized 8x8 IDCT
            C[k1][k2] = (1/4) * sum_j1=0^7 sum_j2=0^7 
                            a[j1][j2] * s[j1] * s[j2] * 
                            cos(pi*j1*(k1+1/2)/8) * 
                            cos(pi*j2*(k2+1/2)/8), 0<=k1<8, 0<=k2<8
                            (s[0] = 1/sqrt(2), sj] = 1, j > 0)
        <case2> Normalized 8x8 DCT
            C[k1][k2] = (1/4) * s[k1] * s[k2] * sum_j1=0^7 sum_j2=0^7 
                            a[j1][j2] * 
                            cos(pi*(j1+1/2)*k1/8) * 
                            cos(pi*(j2+1/2)*k2/8), 0<=k1<8, 0<=k2<8
                            (s[0] = 1/sqrt(2), s[j] = 1, j > 0)
    [usage]
        <case1>
            ddct8x8s(1, a);
        <case2>
            ddct8x8s(-1, a);
    [parameters]
        a[0...7][0...7] :input/output data (double **)
                         output data
                             a[k1][k2] = C[k1][k2], 0<=k1<8, 0<=k2<8
*/


/* Cn_kR = sqrt(2.0/n) * cos(pi/2*k/n) */
/* Cn_kI = sqrt(2.0/n) * sin(pi/2*k/n) */
/* Wn_kR = cos(pi/2*k/n) */
/* Wn_kI = sin(pi/2*k/n) */

#define C8_1R   16070	//0.49039264020161522456<<15
#define C8_1I   3197	//0.09754516100806413392<<15
#define C8_2R   15137	//0.46193976625564337806<<15
#define C8_2I   6270	//0.19134171618254488586<<15
#define C8_3R   13623	//0.41573480615127261854<<15
#define C8_3I   9103	//0.27778511650980111237<<15
#define C8_4R   11586	//0.35355339059327376220<<15
#define W8_4R   23171	//0.70710678118654752440<<15

void ddct8x8s(int isgn, int *block)
{
	int* a = block;

    int j;
    int x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    int xr, xi;
    
    if (isgn < 0) {
        for (j = 0; j <= 7; j++) {
            x0r = a[0] + a[7];
            x1r = a[0] - a[7];
            x0i = a[2] + a[5];
            x1i = a[2] - a[5];
            x2r = a[4] + a[3];
            x3r = a[4] - a[3];
            x2i = a[6] + a[1];
            x3i = a[6] - a[1];
            xr = x0r + x2r;
            xi = x0i + x2i;
            a[0] = (C8_4R * (xr + xi) )>>15;
            a[4] = (C8_4R * (xr - xi) )>>15;
            xr = x0r - x2r;
            xi = x0i - x2i;
            a[2] = (C8_2R * xr - C8_2I * xi )>>15;
            a[6] = (C8_2R * xi + C8_2I * xr )>>15;
            xr = (W8_4R * (x1i - x3i) )>>15;
            x1i = (W8_4R * (x1i + x3i) )>>15;
            x3i = x1i - x3r;
            x1i += x3r;
            x3r = x1r - xr;
            x1r += xr;
            a[1] = (C8_1R * x1r - C8_1I * x1i )>>15;
            a[7] = (C8_1R * x1i + C8_1I * x1r )>>15;
            a[3] = (C8_3R * x3r - C8_3I * x3i )>>15;
            a[5] = (C8_3R * x3i + C8_3I * x3r )>>15;

			a+=8;
        }

		a = block;

        for (j = 0; j <= 7; j++) {
            x0r = a[0*8] + a[7*8];
            x1r = a[0*8] - a[7*8];
            x0i = a[2*8] + a[5*8];
            x1i = a[2*8] - a[5*8];
            x2r = a[4*8] + a[3*8];
            x3r = a[4*8] - a[3*8];
            x2i = a[6*8] + a[1*8];
            x3i = a[6*8] - a[1*8];
            xr = x0r + x2r;
            xi = x0i + x2i;
            a[0*8] = (C8_4R * (xr + xi) )>>15;
            a[4*8] = (C8_4R * (xr - xi) )>>15;
            xr = x0r - x2r;
            xi = x0i - x2i;
            a[2*8] = (C8_2R * xr - C8_2I * xi )>>15;
            a[6*8] = (C8_2R * xi + C8_2I * xr )>>15;
            xr = (W8_4R * (x1i - x3i) )>>15;
            x1i = (W8_4R * (x1i + x3i) )>>15;
            x3i = x1i - x3r;
            x1i += x3r;
            x3r = x1r - xr;
            x1r += xr;
            a[1*8] = (C8_1R * x1r - C8_1I * x1i )>>15;
            a[7*8] = (C8_1R * x1i + C8_1I * x1r )>>15;
            a[3*8] = (C8_3R * x3r - C8_3I * x3i )>>15;
            a[5*8] = (C8_3R * x3i + C8_3I * x3r )>>15;
            
            a++;
        }
    } else {
        for (j = 0; j <= 7; j++) {
            x1r = (C8_1R * a[1] + C8_1I * a[7] )>>15;
            x1i = (C8_1R * a[7] - C8_1I * a[1] )>>15;
            x3r = (C8_3R * a[3] + C8_3I * a[5] )>>15;
            x3i = (C8_3R * a[5] - C8_3I * a[3] )>>15;
            xr = x1r - x3r;
            xi = x1i + x3i;
            x1r += x3r;
            x3i -= x1i;
            x1i = (W8_4R * (xr + xi) )>>15;
            x3r = (W8_4R * (xr - xi) )>>15;
            xr = (C8_2R * a[2] + C8_2I * a[6] )>>15;
            xi = (C8_2R * a[6] - C8_2I * a[2] )>>15;
            x0r = (C8_4R * (a[0] + a[4]) )>>15;
            x0i = (C8_4R * (a[0] - a[4]) )>>15;
            x2r = x0r - xr;
            x2i = x0i - xi;
            x0r += xr;
            x0i += xi;
            a[0] = x0r + x1r;
            a[7] = x0r - x1r;
            a[2] = x0i + x1i;
            a[5] = x0i - x1i;
            a[4] = x2r - x3i;
            a[3] = x2r + x3i;
            a[6] = x2i - x3r;
            a[1] = x2i + x3r;
            
            a+=8;
        }

		a = block;

        for (j = 0; j <= 7; j++) {
            x1r = (C8_1R * a[1*8] + C8_1I * a[7*8] )>>15;
            x1i = (C8_1R * a[7*8] - C8_1I * a[1*8] )>>15;
            x3r = (C8_3R * a[3*8] + C8_3I * a[5*8] )>>15;
            x3i = (C8_3R * a[5*8] - C8_3I * a[3*8] )>>15;
            xr = x1r - x3r;
            xi = x1i + x3i;
            x1r += x3r;
            x3i -= x1i;
            x1i = (W8_4R * (xr + xi) )>>15;
            x3r = (W8_4R * (xr - xi) )>>15;
            xr = (C8_2R * a[2*8] + C8_2I * a[6*8] )>>15;
            xi = (C8_2R * a[6*8] - C8_2I * a[2*8] )>>15;
            x0r = (C8_4R * (a[0*8] + a[4*8]) )>>15;
            x0i = (C8_4R * (a[0*8] - a[4*8]) )>>15;
            x2r = x0r - xr;
            x2i = x0i - xi;
            x0r += xr;
            x0i += xi;
            a[0*8] = x0r + x1r;
            a[7*8] = x0r - x1r;
            a[2*8] = x0i + x1i;
            a[5*8] = x0i - x1i;
            a[4*8] = x2r - x3i;
            a[3*8] = x2r + x3i;
            a[6*8] = x2i - x3r;
            a[1*8] = x2i + x3r;
            
            a++;
        }
    }
}
